#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#define DBG_INFO printf
#define DBG_ERR printf

#define BASE_RIU_PA                 0x1F000000
#define PMSLEEP_BANK                0x000E00
#define PM_SAR_BANK                 0x001400
#define ALBANY1_BANK                0x151500
#define ALBANY2_BANK                0x151600
#define CHIPTOP_BANK                0x101E00
#define PADTOP_BANK                 0x103C00
#define PM_PADTOP_BANK              0x003F00
#define UTMI0_BANK                  0x142100
#define ETH_BANK                    0x003300
#define PM_PADTOP_BANK              0x003F00
#define CHIPTOP_BANK                0x101E00
#define PADTOP_BANK                 0x103C00
#define PADGPIO_BANK                0x103E00
#define PM_SAR_BANK                 0x001400
#define PMSLEEP_BANK                0x000E00
#define PM_GPIO_BANK                0x000F00

#define PAD_GPIO0                   0x3D
#define TEST_GPIO                   61

#define GET_BANK_ADDR(base,bank,offset)   ((base) + ((bank) << 1) + ((offset) << 2))

#define REG_W_WORD(addr,val)              {(*(unsigned short *)(addr)) = (unsigned short)(val);}
#define REG_W_WORD_MASK(addr,val,mask)    {(*(unsigned short *)(addr)) = ((*(unsigned short *)(addr)) & ~(mask)) | ((unsigned short)(val) & (mask));}
#define REG_R_WORD(addr)                  (*(unsigned short *)(addr))
#define REG_R_WORD_MASK(addr,mask)        (*(unsigned short *)(addr)) & (mask))

#define BIT8        0x100

#define BANK_TO_ADDR32(b) (b<<9)
#define REG_ADDR(riu_base,bank,reg_offset) ((riu_base)+BANK_TO_ADDR32(bank)+(reg_offset*4))

typedef struct
{
    unsigned char *virt_addr;
    unsigned char *mmap_base;
    unsigned int mmap_length;
}MmapHandle;

static unsigned int const page_size_mask = 0xFFF;

MmapHandle* devMemMMap(unsigned int phys_addr, unsigned int length)
{
    int fd;
    unsigned int phys_offset = 0;

    fd = open("/dev/mem", O_RDWR|O_SYNC);
    if (fd == -1)
    {
        DBG_ERR("open /dev/mem fail\n");
        return NULL;
    }

    MmapHandle *handle = malloc(sizeof(MmapHandle));
    //phys_offset =(phys_addr & (page_size_mask));
    //phys_addr &= ~(page_size_mask);
    handle->mmap_length = length + phys_offset;
    handle->mmap_base = mmap(NULL, handle->mmap_length, PROT_READ|PROT_WRITE, MAP_SHARED, fd, phys_addr);
    handle->virt_addr = handle->mmap_base + phys_offset;
    DBG_INFO("phys_addr: %#x\n", phys_addr);
    DBG_INFO("virt_addr: %p\n", handle->virt_addr);
    DBG_INFO("phys_offset: %#x\n", phys_offset);

    if (handle->mmap_base == MAP_FAILED)
    {
        DBG_ERR("mmap fail\n");
        close(fd);
        free(handle);
        return NULL;
    }

    close(fd);
    return handle;
}

int devMemUmap(MmapHandle* handle)
{
    int ret = 0;

    ret = munmap(handle->mmap_base, handle->mmap_length);
    if(ret != 0)
    {
        printf("munmap fail\n");
        return ret;
    }
    free(handle);
    return ret;
}


int main(int argc, char *argv[])
{
    unsigned short content;
    int value = 0;
    /* RIU mapping*/
    MmapHandle *riu_base = devMemMMap(BASE_RIU_PA, 0x400000);

    if (!strcmp(argv[1], "input"))
    {
        REG_W_WORD(GET_BANK_ADDR(riu_base->virt_addr, PADGPIO_BANK, PAD_GPIO0), 0x7C);
        content = REG_R_WORD(GET_BANK_ADDR(riu_base->virt_addr, PADGPIO_BANK, PAD_GPIO0));
        value = (content & 0x0001) ? 1 : 0;
        printf("gpio[%d] value = %d\n", TEST_GPIO, value);
    }
    else if (!strcmp(argv[1], "output"))
    {
        if (argc < 3)
        {
            devMemUmap(riu_base);
            return -1;
        }
        value = atoi(argv[2]);
        if (value)
        {
            REG_W_WORD(GET_BANK_ADDR(riu_base->virt_addr, PADGPIO_BANK, PAD_GPIO0), 0x7B);
        }
        else
        {
            REG_W_WORD(GET_BANK_ADDR(riu_base->virt_addr, PADGPIO_BANK, PAD_GPIO0), 0x78);
        }
    }
    else
    {
        printf("unknown commamd\n");
    }

    devMemUmap(riu_base);

    return 0;
}

